@@ -19,7 +19,6 @@
 #undef kmem_free
 #undef mem_alloc
 #undef mem_free
-#undef register
 #endif /* AFS_ALPHA_ENV */
 #include "afsincludes.h"	/* Afs-based standard headers */
 #include "afs/afs_stats.h"	/* statistics */
@@ -27,11 +26,10 @@
 
 extern int cacheDiskType;
 
-
 #ifndef AFS_NOSTATS
 void
 FillStoreStats(int code, int idx, osi_timeval_t *xferStartTime,
-		  afs_size_t bytesToXfer, afs_size_t bytesXferred)
+	       afs_size_t bytesToXfer, afs_size_t bytesXferred)
 {
     struct afs_stats_xferData *xferP;
     osi_timeval_t xferStopTime;
@@ -95,9 +93,8 @@ rxfs_storeMemPrepare(void *r, afs_uint32 size, afs_uint32 *tlen)
     afs_int32 code;
     struct rxfs_storeVariables *v = (struct rxfs_storeVariables *) r;
 
-    *tlen = (size > AFS_LRALLOCSIZ ?  AFS_LRALLOCSIZ : size);
     RX_AFS_GUNLOCK();
-    code = rx_WritevAlloc(v->call, v->tiov, &v->tnio, RX_MAXIOVECS, *tlen);
+    code = rx_WritevAlloc(v->call, v->tiov, &v->tnio, RX_MAXIOVECS, size);
     RX_AFS_GLOCK();
     if (code <= 0) {
 	code = rx_Error(v->call);
@@ -271,7 +268,7 @@ afs_GenericStoreProc(struct storeOps *ops, void *rock,
 {
     struct rxfs_storeVariables *svar = rock;
     afs_uint32 tlen, bytesread, byteswritten;
-    afs_int32 code;
+    afs_int32 code = 0;
     int offset = 0;
     afs_size_t size;
     struct osi_file *tfile;
@@ -315,7 +312,7 @@ afs_GenericStoreProc(struct storeOps *ops, void *rock,
 
 static
 struct storeOps rxfs_storeUfsOps = {
-#if (defined(AFS_SGI_ENV) && !defined(__c99))
+#ifndef HAVE_STRUCT_LABEL_SUPPORT
     rxfs_storeUfsPrepare,
     rxfs_storeUfsRead,
     rxfs_storeUfsWrite,
@@ -342,14 +339,15 @@ struct storeOps rxfs_storeUfsOps = {
 
 static
 struct storeOps rxfs_storeMemOps = {
-#if (defined(AFS_SGI_ENV) && !defined(__c99))
+#ifndef HAVE_STRUCT_LABEL_SUPPORT
     rxfs_storeMemPrepare,
     rxfs_storeMemRead,
     rxfs_storeMemWrite,
     rxfs_storeStatus,
     rxfs_storePadd,
     rxfs_storeClose,
-    rxfs_storeDestroy
+    rxfs_storeDestroy,
+    afs_GenericStoreProc
 #else
     .prepare =	rxfs_storeMemPrepare,
     .read = 	rxfs_storeMemRead,
@@ -357,14 +355,16 @@ struct storeOps rxfs_storeMemOps = {
     .status =	rxfs_storeStatus,
     .padd =	rxfs_storePadd,
     .close = 	rxfs_storeClose,
-    .destroy =	rxfs_storeDestroy
+    .destroy =	rxfs_storeDestroy,
+    .storeproc = afs_GenericStoreProc
 #endif
 };
 
 afs_int32
-rxfs_storeInit(struct vcache *avc, struct afs_conn *tc, afs_size_t base,
-		afs_size_t bytes, afs_size_t length,
-		int sync, struct storeOps **ops, void **rock)
+rxfs_storeInit(struct vcache *avc, struct afs_conn *tc,
+	       struct rx_connection *rxconn, afs_size_t base,
+	       afs_size_t bytes, afs_size_t length,
+	       int sync, struct storeOps **ops, void **rock)
 {
     afs_int32 code;
     struct rxfs_storeVariables *v;
@@ -457,16 +457,13 @@ unsigned int storeallmissing = 0;
  */
 afs_int32
 afs_CacheStoreDCaches(struct vcache *avc, struct dcache **dclist,
-			afs_size_t bytes,
-			afs_hyper_t *anewDV,
-			int *doProcessFS,
-			struct AFSFetchStatus *OutStatus,
-			afs_uint32 nchunks,
-			int nomore,
-			struct storeOps *ops, void *rock)
+		      afs_size_t bytes, afs_hyper_t *anewDV, int *doProcessFS,
+		      struct AFSFetchStatus *OutStatus, afs_uint32 nchunks,
+		      int nomore, struct storeOps *ops, void *rock)
 {
     int *shouldwake = NULL;
     unsigned int i;
+    int stored = 0;
     afs_int32 code = 0;
     afs_size_t bytesXferred;
 
@@ -477,7 +474,6 @@ afs_CacheStoreDCaches(struct vcache *avc, struct dcache **dclist,
     XSTATS_DECLS;
 
     for (i = 0; i < nchunks && !code; i++) {
-	int stored = 0;
 	struct dcache *tdc = dclist[i];
 	afs_int32 size = tdc->f.chunkBytes;
 	if (!tdc) {
@@ -544,6 +540,7 @@ afs_CacheStoreDCaches(struct vcache *avc, struct dcache **dclist,
 
     if (!code) {
 	code = (*ops->close)(rock, OutStatus, doProcessFS);
+	/* if this succeeds, dv has been bumped. */
 	if (*doProcessFS) {
 	    hadd32(*anewDV, 1);
 	}
@@ -551,6 +548,11 @@ afs_CacheStoreDCaches(struct vcache *avc, struct dcache **dclist,
     }
     if (ops)
 	code = (*ops->destroy)(&rock, code);
+
+    /* if we errored, can't trust this. */
+    if (code)
+	*doProcessFS = 0;
+
     return code;
 }
 
@@ -572,10 +574,9 @@ afs_CacheStoreDCaches(struct vcache *avc, struct dcache **dclist,
  */
 int
 afs_CacheStoreVCache(struct dcache **dcList, struct vcache *avc,
-			struct vrequest *areq, int sync,
-			unsigned int minj, unsigned int high,
-			unsigned int moredata,
-			afs_hyper_t *anewDV, afs_size_t *amaxStoredLength)
+		     struct vrequest *areq, int sync, unsigned int minj,
+		     unsigned int high, unsigned int moredata,
+		     afs_hyper_t *anewDV, afs_size_t *amaxStoredLength)
 {
     afs_int32 code = 0;
     struct storeOps *ops;
@@ -588,6 +589,7 @@ afs_CacheStoreVCache(struct dcache **dcList, struct vcache *avc,
     int nomore;
     unsigned int first = 0;
     struct afs_conn *tc;
+    struct rx_connection *rxconn;
 
     for (bytes = 0, j = 0; !code && j <= high; j++) {
 	if (dcList[j]) {
@@ -632,12 +634,12 @@ afs_CacheStoreVCache(struct dcache **dcList, struct vcache *avc,
 		       ICL_HANDLE_OFFSET(length));
 
 	    do {
-	        tc = afs_Conn(&avc->f.fid, areq, 0);
+		tc = afs_Conn(&avc->f.fid, areq, 0, &rxconn);
 
 #ifdef AFS_64BIT_CLIENT
 	      restart:
 #endif
-		code = rxfs_storeInit(avc, tc, base, bytes, length,
+		code = rxfs_storeInit(avc, tc, rxconn, base, bytes, length,
 				      sync, &ops, &rock);
 		if ( !code ) {
 		    code = afs_CacheStoreDCaches(avc, dclist, bytes, anewDV,
@@ -652,7 +654,7 @@ afs_CacheStoreVCache(struct dcache **dcList, struct vcache *avc,
 		}
 #endif /* AFS_64BIT_CLIENT */
 	    } while (afs_Analyze
-		     (tc, code, &avc->f.fid, areq,
+		     (tc, rxconn, code, &avc->f.fid, areq,
 		      AFS_STATS_FS_RPCIDX_STOREDATA, SHARED_LOCK,
 		      NULL));
 
@@ -769,9 +771,8 @@ rxfs_fetchMemRead(void *r, afs_uint32 tlen, afs_uint32 *bytesread)
 
 
 afs_int32
-rxfs_fetchMemWrite(void *r, struct osi_file *fP,
-			afs_uint32 offset, afs_uint32 tlen,
-			afs_uint32 *byteswritten)
+rxfs_fetchMemWrite(void *r, struct osi_file *fP, afs_uint32 offset,
+		   afs_uint32 tlen, afs_uint32 *byteswritten)
 {
     afs_int32 code;
     struct rxfs_fetchVariables *v = (struct rxfs_fetchVariables *)r;
@@ -786,9 +787,8 @@ rxfs_fetchMemWrite(void *r, struct osi_file *fP,
 }
 
 afs_int32
-rxfs_fetchUfsWrite(void *r, struct osi_file *fP,
-			afs_uint32 offset, afs_uint32 tlen,
-			afs_uint32 *byteswritten)
+rxfs_fetchUfsWrite(void *r, struct osi_file *fP, afs_uint32 offset,
+		   afs_uint32 tlen, afs_uint32 *byteswritten)
 {
     afs_int32 code;
     struct rxfs_fetchVariables *v = (struct rxfs_fetchVariables *)r;
@@ -804,7 +804,7 @@ rxfs_fetchUfsWrite(void *r, struct osi_file *fP,
 
 afs_int32
 rxfs_fetchClose(void *r, struct vcache *avc, struct dcache * adc,
-					struct afs_FetchOutput *o)
+		struct afs_FetchOutput *o)
 {
     afs_int32 code, code1 = 0;
     struct rxfs_fetchVariables *v = (struct rxfs_fetchVariables *)r;
@@ -903,9 +903,10 @@ struct fetchOps rxfs_fetchMemOps = {
 };
 
 afs_int32
-rxfs_fetchInit(struct afs_conn *tc, struct vcache *avc, afs_offs_t base,
-		afs_uint32 size, afs_int32 *alength, struct dcache *adc,
-		struct osi_file *fP, struct fetchOps **ops, void **rock)
+rxfs_fetchInit(struct afs_conn *tc, struct rx_connection *rxconn,
+	       struct vcache *avc, afs_offs_t base,
+	       afs_uint32 size, afs_int32 *alength, struct dcache *adc,
+	       struct osi_file *fP, struct fetchOps **ops, void **rock)
 {
     struct rxfs_fetchVariables *v;
     int code = 0, code1 = 0;
@@ -914,13 +915,14 @@ rxfs_fetchInit(struct afs_conn *tc, struct vcache *avc, afs_offs_t base,
 #endif
     afs_uint32 length, bytes;
 
-    v = (struct rxfs_fetchVariables *) osi_AllocSmallSpace(sizeof(struct rxfs_fetchVariables));
+    v = (struct rxfs_fetchVariables *)
+	    osi_AllocSmallSpace(sizeof(struct rxfs_fetchVariables));
     if (!v)
         osi_Panic("rxfs_fetchInit: osi_AllocSmallSpace returned NULL\n");
     memset(v, 0, sizeof(struct rxfs_fetchVariables));
 
     RX_AFS_GUNLOCK();
-    v->call = rx_NewCall(tc->id);
+    v->call = rx_NewCall(rxconn);
     RX_AFS_GLOCK();
     if (v->call) {
 #ifdef AFS_64BIT_CLIENT
@@ -928,8 +930,9 @@ rxfs_fetchInit(struct afs_conn *tc, struct vcache *avc, afs_offs_t base,
 	if (!afs_serverHasNo64Bit(tc)) {
 	    afs_uint64 llbytes = size;
 	    RX_AFS_GUNLOCK();
-	    code = StartRXAFS_FetchData64(v->call, (struct AFSFid *) &avc->f.fid.Fid,
-					       base, llbytes);
+	    code = StartRXAFS_FetchData64(v->call,
+					  (struct AFSFid *) &avc->f.fid.Fid,
+					  base, llbytes);
 	    if (code != 0) {
 		RX_AFS_GLOCK();
 		afs_Trace2(afs_iclSetp, CM_TRACE_FETCH64CODE,
@@ -956,7 +959,7 @@ rxfs_fetchInit(struct afs_conn *tc, struct vcache *avc, afs_offs_t base,
 		pos = base;
 		RX_AFS_GUNLOCK();
 		if (!v->call)
-		    v->call = rx_NewCall(tc->id);
+		    v->call = rx_NewCall(rxconn);
 		code =
 		    StartRXAFS_FetchData(
 		    		v->call, (struct AFSFid*)&avc->f.fid.Fid,
@@ -1061,6 +1064,7 @@ rxfs_fetchInit(struct afs_conn *tc, struct vcache *avc, afs_offs_t base,
  *	that more has arrived.
  *
  * \param tc Ptr to the Rx connection structure.
+ * \param rxconn Ptr to the Rx connection structure.
  * \param fP File descriptor for the cache file.
  * \param base Base offset to fetch.
  * \param adc Ptr to the dcache entry for the file, write-locked.
@@ -1071,9 +1075,10 @@ rxfs_fetchInit(struct afs_conn *tc, struct vcache *avc, afs_offs_t base,
  * \note Environment: Nothing interesting.
  */
 int
-afs_CacheFetchProc(struct afs_conn *tc, struct osi_file *fP, afs_size_t base,
-		    struct dcache *adc, struct vcache *avc, afs_int32 size,
-		    struct afs_FetchOutput *tsmall)
+afs_CacheFetchProc(struct afs_conn *tc, struct rx_connection *rxconn,
+		   struct osi_file *fP, afs_size_t base,
+		   struct dcache *adc, struct vcache *avc, afs_int32 size,
+		   struct afs_FetchOutput *tsmall)
 {
     afs_int32 code;
     afs_int32 length;
@@ -1100,7 +1105,7 @@ afs_CacheFetchProc(struct afs_conn *tc, struct osi_file *fP, afs_size_t base,
      * adc->lock(W)
      */
     code = rxfs_fetchInit(
-		tc, avc, base, size, &length, adc, fP, &ops, &rock);
+	tc, rxconn, avc, base, size, &length, adc, fP, &ops, &rock);
 
 #ifndef AFS_NOSTATS
     osi_GetuTime(&xferStartTime);
